"""Utility functions for IVOA clients."""

import os

from pyvo.dal import SIA2Service
from pyvo.dal.adhoc import DatalinkResults
from pyvo.dal.sia2 import ObsCoreRecord

from .utils import get_pyvo_auth, get_service_url


def get_datalink_result(result: ObsCoreRecord) -> DatalinkResults:
    """Return the datalink part of a result."""
    return DatalinkResults.from_result_url(
        result.getdataurl(), session=get_pyvo_auth()
    )


def get_siav2_service(data_release: str) -> SIA2Service:
    """Construct an `SIA2Service` client."""
    # data_release determines the Data release, as we may have different
    # releases being served from the same server.

    label = os.getenv("RSP_SITE_TYPE", None)

    # The label variable here corresponds to RSP "kind" (i.e. "telescope",
    # "science", "staff")

    if label not in {"science", "staff"}:
        if label is None:
            raise ValueError("RSP_SITE_TYPE environment variable is not set")
        raise ValueError(f"{label} data not available at your location")

    sia_url = get_service_url(f"sia/{data_release}")
    if not sia_url:
        raise RuntimeError(
            f"Failed to determine service URL for data release: {data_release}"
        )

    session = get_pyvo_auth()
    if session:
        session.add_security_method_for_url(sia_url + "/query", "lsst-token")

    return SIA2Service(sia_url, session=session)



"""A tool to set up user directories to hold what will be the default
page for "science" sites, to allow those sites to load a tutorial
document on startup.

It has two phases: first, it creates copies of the landing page and
supporting files, and second, if the user default for opening the file
is not the Markdown Viewer already, it writes out configuration for
that.

The files must be somewhere that the Notebook container can open them, and
must be writeable files (not symlinks to a read-only target) if the
Save-All or Save-And-Quite functionality is to work.

The tool is expected to be running in the context of the current user,
as part of an initContainer running after the user home directories
are provisioned, but before the user lab container begins to start.
"""

import json
import shutil
from typing import Any, Self

from .exceptions import (
    DestinationError,
    DestinationIsDirectoryError,
    PrecheckError,
)
from .input import ProvisionerInput, input_from_env


class Provisioner:
    """Ensure user environment is ready to open tutorial landing page.

    Parameters
    ----------
    inp
        Input document specifying document source, user home, destination
        directory, and whether to enable debugging.
    """

    def __init__(self, inp: ProvisionerInput) -> None:
        self._debug = inp.debug
        self._source_files = inp.source_files
        self._home_dir = inp.home_dir
        self._dest_dir = inp.dest_dir

    def _precheck(self) -> None:
        for sf in self._source_files:
            if not sf.is_file():
                raise PrecheckError(f"Source file {sf} is not a file")
        if not self._home_dir.is_dir():
            raise PrecheckError(
                "Home directory {self._home_dir} is not a directory"
            )

    def _provision_tutorial_directories(self) -> None:
        self._dest_dir.mkdir(mode=0o755, exist_ok=True, parents=True)

    def _copy_files(self) -> None:
        for src in self._source_files:
            dest = self._dest_dir / src.name
            if dest.exists(follow_symlinks=False):
                if dest.is_symlink() or dest.is_file():
                    # Turns out a symlink to a read-only file isn't going
                    # to work.
                    #
                    # Remove and recopy.  It's probably imperceptibly slower
                    # than trying to be clever about it.
                    dest.unlink()
                elif dest.is_dir():
                    raise DestinationIsDirectoryError(str(dest))
                else:
                    # It's...a device, or a named pipe, or ... something?
                    raise DestinationError(str(dest))
            shutil.copy(src, dest)

    def _edit_settings(self) -> None:
        settings_dir = (
            self._home_dir
            / ".jupyter"
            / "lab"
            / "user-settings"
            / "@jupyterlab"
            / "docmanager-extension"
        )
        settings_dir.mkdir(exist_ok=True, parents=True)
        settings_file = settings_dir / "plugin.jupyterlab-settings"
        settings: dict[str, Any] = {}
        write = False
        if settings_file.exists():
            # It's supposed to be a JSON doc.
            settings = json.loads(settings_file.read_text())
        if "defaultViewers" not in settings:
            settings["defaultViewers"] = {}
        if "markdown" not in settings["defaultViewers"]:
            # If the user set it to something else, leave it alone.
            settings["defaultViewers"]["markdown"] = "Markdown Preview"
            write = True
        if write:
            settings_file.write_text(
                json.dumps(settings, sort_keys=True, indent=2)
            )

    def go(self) -> None:
        """Do the deed."""
        self._precheck()
        self._provision_tutorial_directories()
        self._copy_files()
        self._edit_settings()

    @classmethod
    def from_env(cls) -> Self:
        """Create provisioner from environment."""
        inp = input_from_env()
        return cls(inp)



"""Pytest configuration and fixtures."""

import contextlib
import os
from collections.abc import Iterator
from pathlib import Path
from shutil import copytree
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pytest

from lsst.rsp.startup.storage.command import Command

# Things for startup/labrunner


@pytest.fixture
def _rsp_paths(monkeypatch: pytest.MonkeyPatch) -> Iterator[None]:
    # For each of these, we want to cover both the "from ..constants import"
    # and the "import lsst.rsp.constants" case.
    with patch(
        "lsst.rsp.startup.services.labrunner.labrunner.ETC_PATH",
        (Path(__file__).parent / "support" / "files" / "etc"),
    ):
        with patch(
            "lsst.rsp.startup.constants.ETC_PATH",
            (Path(__file__).parent / "support" / "files" / "etc"),
        ):
            yield


@pytest.fixture
def _rsp_env(
    _rsp_paths: None, monkeypatch: pytest.MonkeyPatch
) -> Iterator[None]:
    file_dir = Path(__file__).parent / "support" / "files"
    template = file_dir / "homedir"
    monkeypatch.setenv(
        "NUBLADO_RUNTIME_MOUNTS_DIR", str(file_dir / "etc" / "nublado")
    )
    monkeypatch.setenv(
        "JUPYTERLAB_CONFIG_DIR",
        str(file_dir / "jupyterlab"),
    )
    with contextlib.suppress(KeyError):
        monkeypatch.delenv("TMPDIR")
        monkeypatch.delenv("DAF_BUTLER_CACHE_DIRECTORY")
    with TemporaryDirectory() as fake_root:
        t_home = Path(fake_root) / "home"
        t_home.mkdir()
        homedir = str(t_home)
        monkeypatch.setenv("HOME", homedir)
        monkeypatch.setenv("USER", "hambone")
        monkeypatch.setenv("JUPYTERHUB_BASE_URL", "/nb/")
        copytree(
            template,
            homedir,
            dirs_exist_ok=True,
            symlinks=True,
        )
        t_scratch = Path(fake_root) / "scratch"
        t_scratch.mkdir()
        monkeypatch.setenv("SCRATCH_PATH", str(t_scratch))
        yield


@pytest.fixture
def git_repo() -> Iterator[Path]:
    with TemporaryDirectory() as repo_str:
        pwd = Path.cwd()
        repo = Path(repo_str)
        os.chdir(repo)
        cmd = Command()
        cmd.run("git", "init")
        (repo / "README.md").write_text("# Test Repo\n")
        cmd.run("git", "config", "user.email", "hambone@opera.borphee.quendor")
        cmd.run("git", "config", "user.name", "Hambone")
        cmd.run("git", "config", "init.defaultBranch", "main")
        cmd.run("git", "checkout", "-b", "main")
        cmd.run("git", "add", "README.md")
        cmd.run("git", "commit", "-am", "Initial Commit")
        os.chdir(pwd)
        yield Path(repo)


# Things for startup/landing_page.


@pytest.fixture(scope="session")
def monkeysession() -> Iterator[pytest.MonkeyPatch]:
    """MonkeyPatch, but session-scoped."""
    mpatch = pytest.MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(scope="session")
def _init_container_fake_root(
    monkeysession: pytest.MonkeyPatch,
) -> Iterator[None]:
    with TemporaryDirectory() as td:
        contents = {
            "hello.txt": "Hello, world!\n",
            "goodbye.txt": "Goodbye, cruel world.\n",
        }
        tutorial_directory = Path(td) / "tutorials"
        tutorial_directory.mkdir(parents=True)
        for fn, text in contents.items():
            out_file = tutorial_directory / fn
            out_file.write_text(text)
        home_directory = Path(td) / "home" / "gregorsamsa"
        home_directory.mkdir(parents=True)

        monkeysession.setenv("NUBLADO_HOME", str(home_directory))
        monkeysession.setenv(
            "CST_LANDING_PAGE_SRC_DIR", str(tutorial_directory)
        )
        monkeysession.setenv("CST_LANDING_PAGE_TGR_DIR", "notebooks/tutorials")
        monkeysession.setenv(
            "CST_LANDING_PAGE_FILES", "hello.txt", "goodbye.txt"
        )
        yield



"""Exceptions for the LabRunner startup service."""

from __future__ import annotations

import errno
import os
import subprocess
from collections.abc import Iterable
from enum import Enum
from shlex import join
from typing import Any, Self

__all__ = [
    "CommandFailedError",
    "CommandTimedOutError",
    "RSPErrorCode",
    "RSPStartupError",
]


class CommandFailedError(Exception):
    """Execution of a command failed.

    Parameters
    ----------
    args
        Command (args[0]) and arguments to that command.
    exc
        Exception reporting the failure.

    Attributes
    ----------
    stdout
        Standard output from the failed command.
    stderr
        Standard error from the failed command.
    """

    def __init__(
        self,
        args: Iterable[str],
        exc: subprocess.CalledProcessError,
    ) -> None:
        args_str = join(args)
        msg = f"'{args_str}' failed with status {exc.returncode}"
        super().__init__(msg)
        self.stdout = exc.stdout
        self.stderr = exc.stderr


class CommandTimedOutError(Exception):
    """Execution of a command failed.

    Parameters
    ----------
    args
        Command (args[0]) and arguments to that command.
    exc
        Exception reporting the failure.

    Attributes
    ----------
    stdout
        Standard output from the failed command.
    stderr
        Standard error from the failed command.
    """

    def __init__(
        self,
        args: Iterable[str],
        exc: subprocess.TimeoutExpired,
    ) -> None:
        args_str = join(args)
        msg = f"'{args_str}' timed out after {exc.timeout}s"
        super().__init__(msg)
        self.stdout = exc.stdout
        self.stderr = exc.stderr


# These are new errors, which are structured like OSError, but aren't.
# OSError's errno tops out at 106 as of Python 3.12 on x64 Linux, so we will
# start at 200 to give that some expansion room.


class RSPErrorCode(Enum):
    """New Error codes for RSP Startup."""

    EBADENV = 200
    EUNKNOWN = 201


# Used internally to populate our RSPStartupErrors
_rsp_errors: dict[int, dict[str, str | int]] = {
    RSPErrorCode.EBADENV.value: {
        "errorcode": "EBADENV",  # Bad environment variable
        "strerror": "Missing environment variable",
    },
    RSPErrorCode.EUNKNOWN.value: {
        "errorcode": "EUNKNOWN",  # Unknown error
        "strerror": f"Unknown error {RSPErrorCode.EUNKNOWN.value}",
    },
}


class RSPStartupError(OSError):
    """RSPStartupError is a subclass of OSError that is designed to be
    more portable than the standard OSError, since we are throwing it
    to a client that could, potentially, be running on a different
    architecture or OS, and whose numeric error codes might therefore
    not match (e.g. `EDQUOT` is 69 under MacOS aarch64, but 122 for
    Linux x64).

    This also gives us the opportunity to set the `filename` parameter
    to, for instance, indicate a missing environment variable.
    """

    # Additional errors we're defining, not present in
    # OSError
    #
    # For Python 3.12 on x64 Linux, at least, errno.errorcode has a greatest
    # value of 106.  So we're going to start at 200 for our custom errors.

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        errnum = self.errno
        # See if it's one of our custom errors or unknown.
        if errnum is None:  # Just to keep mypy happy; can't happen.
            errnum = RSPErrorCode.EUNKNOWN.value
        vals = [x.value for x in RSPErrorCode]
        ec = errno.errorcode.get(errnum)
        if ec:
            self.errorcode = ec
            return
        if errnum not in vals:
            # It's not one of ours, and it's not standard.
            # That makes it unknown.
            # We say 201 the fancy way, in case we ever need to renumber.
            errnum = RSPErrorCode.EUNKNOWN.value
            self.errno = errnum
        errnum = self.errno  # Maybe we reset it to EUNKNOWN
        if errnum is None:  # Just to keep mypy happy; can't happen.
            errnum = RSPErrorCode.EUNKNOWN.value
        # We know this won't be a KeyError
        rsp_e = _rsp_errors[errnum]
        if len(args) > 1:
            strerror = args[1]
        self.strerror = strerror or str(rsp_e["strerror"])
        self.errorcode = str(rsp_e["errorcode"])
        if len(args) > 2:
            self.filename = args[2]
        # Assume winerror doesn't exist--if we're ever running on
        # Windows, we need to revisit.
        if len(args) > 3:
            self.filename2 = args[3]

    @classmethod
    def from_os_error(cls, exc: OSError) -> Self:
        """Create one of these from an underlying OSError exception."""
        # filename2 will probably never be set in the RSP startup use case.
        errnum = exc.errno or RSPErrorCode.EUNKNOWN.value
        strerror = (
            exc.strerror or os.strerror(errnum) or f"Unknown error {errnum}"
        )
        winerror = ""  # Change if we ever need windows support
        return cls(errnum, strerror, exc.filename, winerror, exc.filename2)



"""Entry point for CST tutorial provisioner."""

import logging

from .provisioner import Provisioner

__all__ = ["main"]


def main() -> None:
    """Entry point for provisioner.

    Environment variables ``NUBLADO_HOME`` must be set.  This is
    Nublado-controller specific.

    The other three can come from the user environment or be defaulted.

    We never want to raise an exception from here: if the container fails,
    we still want to start the user lab, but the experience might be
    suboptimal.
    """
    logger = logging.getLogger("cst_tutorial_provisioner")
    logging.basicConfig()  # Not great but it'll do.
    try:
        provisioner = Provisioner.from_env()
        provisioner.go()
    except Exception:
        logger.exception("Provisioner failed")



"""Convenience class and function for handling provisioner inputs."""

import os
from dataclasses import dataclass
from pathlib import Path


@dataclass
class ProvisionerInput:
    """Convenience class for handling provisioner inputs."""

    source_files: list[Path]
    home_dir: Path
    dest_dir: Path
    debug: bool = False


def input_from_env() -> ProvisionerInput:
    """Construct input from environment and defaults."""
    debug = bool(os.getenv("DEBUG", ""))
    home_dir = Path(os.getenv("NUBLADO_HOME", "/nonexistent"))
    source_dir = Path(
        os.getenv(
            "CST_LANDING_PAGE_SRC_DIR",
            "/rubin/cst_repos/tutorial-notebooks-data/welcome",
        )
    )
    target_dir = Path(
        os.getenv("CST_LANDING_PAGE_TGT_DIR", "notebooks/tutorials")
    )
    filelist_str = os.getenv(
        "CST_LANDING_PAGE_FILES", "welcome.md,logo_for_header.png"
    )
    filelist_list = filelist_str.split(",")
    source_files = [source_dir / x for x in filelist_list]
    dest_dir = home_dir / target_dir

    return ProvisionerInput(
        source_files=source_files,
        home_dir=home_dir,
        dest_dir=dest_dir,
        debug=debug,
    )



"""Exceptions for the CST landing page provisioner."""


class PrecheckError(Exception):
    """Initial input-files-and-home-directory sanity check failed."""


class DestinationError(Exception):
    """The destination target exists and cannot be replaced safely."""


class DestinationIsDirectoryError(DestinationError):
    """The destination target exists and is a directory."""



"""Tests for RSPStartupErrror."""

import errno
import os

from lsst.rsp.startup.exceptions import RSPStartupError


def test_bad_environment() -> None:
    assert "NONEXISTENT_ENV_VAR" not in os.environ
    exc = RSPStartupError(200, None, "NONEXISTENT_ENV_VAR")
    assert exc.errno == 200
    assert exc.strerror == "Missing environment variable"
    assert exc.filename == "NONEXISTENT_ENV_VAR"
    assert exc.errorcode == "EBADENV"


def test_from_oserror() -> None:
    osexc = OSError(errno.EDQUOT, None, "/fake/huge-file")
    exc = RSPStartupError.from_os_error(osexc)
    assert exc.errno == errno.EDQUOT
    assert exc.strerror == os.strerror(errno.EDQUOT)
    assert exc.filename == "/fake/huge-file"
    assert exc.errorcode == "EDQUOT"



"""RSP Lab launcher."""

import configparser
import contextlib
import datetime
import hashlib
import json
import os
import pwd
import shutil
import sys
import time
from pathlib import Path
from textwrap import dedent
from typing import Any
from urllib.parse import parse_qsl, urlparse

import structlog

from .... import get_access_token, get_digest
from ....utils import get_jupyterlab_config_dir, get_runtime_mounts_dir
from ...constants import (
    APP_NAME,
    ETC_PATH,
    MAX_NUMBER_OUTPUTS,
    PREVIOUS_LOGGING_CHECKSUMS,
)
from ...exceptions import RSPErrorCode, RSPStartupError
from ...models.noninteractive import NonInteractiveExecutor
from ...storage.command import Command
from ...storage.logging import configure_logging

__all__ = ["LabRunner"]


class LabRunner:
    """Class to start JupyterLab using the environment supplied by
    JupyterHub and the Nublado controller.

    This environment is very Rubin-specific and opinionated, and will
    likely not work for anyone else's science platform.

    If that's you, use this for inspiration, but don't expect this to
    work out of the box.
    """

    def __init__(self) -> None:
        # We start with a copy of our own environment
        self._env = os.environ.copy()
        self._debug = bool(self._env.get("DEBUG", ""))
        configure_logging(debug=self._debug)
        self._logger = structlog.get_logger(APP_NAME)
        self._broken = False
        for req_env in ("JUPYTERHUB_BASE_URL", "HOME"):
            if req_env not in self._env:
                exc = RSPStartupError("EBADENV", None, req_env)
                self._set_abnormal_startup(exc)
        # If no home, use /tmp?  It won't work but at least if we create
        # stuff there it will be harmless, and the user will get a message
        # indicating what's wrong.
        self._home = Path(self._env.get("HOME", "/tmp"))
        self._cmd = Command(ignore_fail=True, logger=self._logger)

    def go(self) -> None:
        """Start the user lab."""
        # If the user somehow manages to screw up their local environment
        # so badly that Jupyterlab won't even start, we will have to
        # bail them out on the fileserver end.  Since Jupyter Lab is in
        # its own venv, which is not writeable by the user, this should
        # require quite a bit of creativity.
        try:
            self._relocate_user_environment_if_requested()
            self._configure_env()
        except OSError as exc:
            self._set_abnormal_startup(exc)

        # Clean up stale cache, check for writeability, try to free some
        # space if necessary.  This stage will manage its own abnormality,
        # since it tries to take some corrective action.
        self._tidy_homedir()

        # If everything seems OK so far, copy files into the user's home
        # space and set up git-lfs.

        if not self._broken:
            try:
                self._copy_files_to_user_homedir()
                self._setup_git()
            except OSError as exc:
                self._set_abnormal_startup(exc)

        # Decide between interactive and noninteractive start, do
        # things that change between those two, and launch the Lab
        self._launch()

    def _set_abnormal_startup(self, exc: OSError) -> None:
        """Take an OSError, convert it into an RSPStartupError if necessary,
        and then set the env variables that rsp-jupyter-extensions will use
        to report the error to the user at Lab startup.
        """
        self._broken = True
        if not isinstance(exc, RSPStartupError):
            # This will also catch the EUNKNOWN case
            new_exc = RSPStartupError.from_os_error(exc)
        else:
            new_exc = exc

        self._env["ABNORMAL_STARTUP"] = "TRUE"
        self._env["ABNORMAL_STARTUP_ERRNO"] = str(new_exc.errno)
        self._env["ABNORMAL_STARTUP_STRERROR"] = (
            # Mypy didn't know the above caught the EUNKNOWN case.
            new_exc.strerror
            or os.strerror(int(new_exc.errno or RSPErrorCode.EUNKNOWN.value))
            or f"Unknown error {new_exc.errno}"
        )
        self._env["ABNORMAL_STARTUP_ERRORCODE"] = new_exc.errorcode
        self._env["ABNORMAL_STARTUP_MESSAGE"] = str(new_exc)
        msg = f"Abnormal RSP startup set with exception {new_exc!s}"
        self._logger.error(msg)

    def _clear_abnormal_startup(self) -> None:
        for e in ("", "_ERRNO", "_STRERROR", "_MESSAGE", "_ERRORCODE"):
            del self._env[f"ABNORMAL_STARTUP{e}"]
            self._broken = False
            self._logger.info("Cleared abnormal startup condition")

    def _relocate_user_environment_if_requested(self) -> None:
        if not self._env.get("RESET_USER_ENV", ""):
            return
        self._logger.debug("User environment relocation requested")
        now = datetime.datetime.now(datetime.UTC).strftime("%Y%m%d%H%M%S")
        reloc = self._home / f".user_env.{now}"
        for candidate in ("cache", "conda", "eups", "local", "jupyter"):
            c_path = self._home / f".{candidate}"
            if c_path.is_dir():
                if not reloc.is_dir():
                    reloc.mkdir()
                tgt = reloc / candidate
                self._logger.debug(f"Moving {c_path.name} to {tgt.name}")
                shutil.move(c_path, tgt)
        u_setups = self._home / "notebooks" / ".user_setups"
        if u_setups.is_file():
            tgt = reloc / "notebooks" / "user_setups"
            tgt.parent.mkdir()
            self._logger.debug(f"Moving {u_setups.name} to {tgt}")
            shutil.move(u_setups, tgt)

    #
    # Next up, a big block of setting up our subprocess environment.
    #
    def _configure_env(self) -> None:
        self._logger.debug("Configuring environment for JupyterLab process")
        self._set_user()
        self._set_tmpdir_if_scratch_available()
        self._set_butler_cache()
        self._set_cpu_variables()
        self._set_image_digest()
        self._expand_panda_tilde()
        self._set_firefly_variables()
        self._force_jupyter_prefer_env_path_false()
        self._set_butler_credential_variables()
        self._logger.debug("Lab process environment", env=self._env)

    def _set_user(self) -> None:
        self._logger.debug("Determining user name")
        user = self._env.get("USER", "")
        if not user:
            user = pwd.getpwuid(os.getuid()).pw_name
            if not user:
                raise RSPStartupError(RSPErrorCode.EBADENV, None, "USER")
            self._env["USER"] = user

    def _check_user_scratch_subdir(self, path: Path) -> Path | None:
        # This is very Rubin specific.  We generally have a large
        # world-writable filesystem in a scratch path.
        #
        # Given a path we will test that SCRATCH_PATH/user/path can be
        # created as a writable directory (or that it already exists
        # as a writable directory).  If it can be (or is), we return the
        # whole path, and if not, we return None.  If we can set it,
        # we also set the SCRATCH_DIR environment variable to point to it.
        #
        # This will only be readable by the user; they can chmod() it if
        # they want to share, but for TMPDIR and DAF_BUTLER_CACHE_DIRECTORY
        # they probably should not.  The mode will not be reset if the
        # directory already exists and is writeable

        scratch_path = Path(os.getenv("SCRATCH_PATH") or "/scratch")

        if not scratch_path.is_dir():
            self._logger.debug(
                # Debug only: not having /scratch is reasonable.
                f"{scratch_path} is not a directory."
            )
            return None
        user = self._env.get("USER", "")
        if not user:
            self._logger.warning("Could not determine user from environment")
            return None
        schema = self._env.get("HOMEDIR_SCHEMA", "username")
        user_scratch_dir = scratch_path / user
        # This is pretty ad-hoc, but USDF uses the first letter in the
        # username for both home and scratch
        if schema == "initialThenUsername":
            user_scratch_dir = scratch_path / user[0] / user
        user_scratch_path = user_scratch_dir / path
        try:
            user_scratch_path.mkdir(parents=True, exist_ok=True, mode=0o700)
        except OSError as exc:
            self._logger.warning(
                f"Could not create directory at {user_scratch_path!s}: {exc}"
            )
            return None
        if not os.access(user_scratch_path, os.W_OK):
            self._logger.warning(f"Unable to write to {user_scratch_path!s}")
            return None
        self._logger.debug(f"Using user scratch path {user_scratch_path!s}")
        # Set user-specific top dir as SCRATCH_DIR
        self._env["SCRATCH_DIR"] = f"{user_scratch_dir!s}"
        return user_scratch_path

    def _set_tmpdir_if_scratch_available(self) -> None:
        # Assuming that TMPDIR is not already set (e.g. by the spawner),
        # we will try to create <scratch_path>/<user>/tmp and ensure it is a
        # writeable directory, and if it is, TMPDIR will be repointed to it.
        # This will then reduce our ephemeral storage issues, which have
        # caused mass pod eviction and destruction of the prepull cache.
        #
        # In our tests at the IDF, on a 2CPU/8GiB "Medium", TMPDIR on
        # /scratch (NFS) is about 15% slower than on local ephemeral storage.
        self._logger.debug("Resetting TMPDIR if scratch storage available")
        tmpdir = self._env.get("TMPDIR", "")
        if tmpdir:
            self._logger.debug(f"Not setting TMPDIR: already set to {tmpdir}")
            return
        temp_path = self._check_user_scratch_subdir(Path("tmp"))
        if temp_path:
            self._env["TMPDIR"] = str(temp_path)
            self._logger.debug(f"Set TMPDIR to {temp_path!s}")
        else:
            self._logger.debug("Did not set TMPDIR")

    def _set_butler_cache(self) -> None:
        # This is basically the same story as TMPDIR.
        env_v = "DAF_BUTLER_CACHE_DIRECTORY"
        dbcd = self._env.get(env_v, "")
        if dbcd:
            self._logger.debug(
                f"Not setting {env_v}: already set to" f" {dbcd}"
            )
            return
        temp_path = self._check_user_scratch_subdir(Path("butler_cache"))
        if temp_path:
            self._env[env_v] = str(temp_path)
            self._logger.debug(f"Set {env_v} to {temp_path!s}")
            return
        # In any sane RSP environment, /tmp will not be shared (it will
        # be either tmpfs or on ephemeral storage, and in any case not
        # visible beyond its own pod), so we are not actually using a risky
        # shared directory.
        self._env[env_v] = "/tmp/butler_cache"

    def _set_cpu_variables(self) -> None:
        self._logger.debug("Setting CPU threading variables")
        try:
            cpu_limit = int(float(self._env.get("CPU_LIMIT", "1")))
        except ValueError:
            cpu_limit = 1
        if cpu_limit < 1:
            cpu_limit = 1
        cpu_limit_str = str(cpu_limit)
        for vname in (
            "CPU_LIMIT",
            "CPU_COUNT",
            "GOTO_NUM_THREADS",
            "MKL_DOMAIN_NUM_THREADS",
            "MPI_NUM_THREADS",
            "NUMEXPR_NUM_THREADS",
            "NUMEXPR_MAX_THREADS",
            "OMP_NUM_THREADS",
            "OPENBLAS_NUM_THREADS",
            "RAYON_NUM_THREADS",
        ):
            self._env[vname] = cpu_limit_str
            self._logger.debug(f"Set '{vname}' -> '{cpu_limit_str}'")

    def _set_image_digest(self) -> None:
        self._logger.debug("Setting image digest if available")
        # get_digest() is already a helper function in our parent package.
        digest = get_digest()
        if digest:
            self._logger.debug(f"Set image digest to '{digest}'")
            self._env["IMAGE_DIGEST"] = digest
        else:
            self._logger.debug("Could not get image digest")

    def _expand_panda_tilde(self) -> None:
        self._logger.debug("Expanding tilde in PANDA_CONFIG_ROOT, if needed")
        if "PANDA_CONFIG_ROOT" in self._env:
            # We've already been through set_user(), so USER must be set.
            username = self._env["USER"]
            path = Path(self._env["PANDA_CONFIG_ROOT"])
            path_parts = path.parts
            if path_parts[0] in ("~", f"~{username}"):
                new_path = Path(self._home, *path_parts[1:])
                self._logger.debug(
                    f"Replacing PANDA_CONFIG_ROOT '{path!s}'"
                    f"with '{new_path!s}'"
                )
                self._env["PANDA_CONFIG_ROOT"] = str(new_path)
            elif path_parts[0].startswith("~"):
                self._logger.warning(f"Cannot expand tilde in '{path!s}'")

    def _set_firefly_variables(self) -> None:
        self._logger.debug("Setting firefly variables")
        firefly_route = self._env.get("FIREFLY_ROUTE", "/firefly")
        ext_url = self._env.get(
            "EXTERNAL_INSTANCE_URL", "https://localhost:8888"
        )
        self._env["FIREFLY_URL"] = (
            f"{ext_url.strip('/')}/{firefly_route.lstrip('/')}"
        )
        self._logger.debug(f"Firefly URL -> '{self._env['FIREFLY_URL']}'")

    def _force_jupyter_prefer_env_path_false(self) -> None:
        # cf https://discourse.jupyter.org/t/jupyter-paths-priority-order/7771
        # and https://jupyter-core.readthedocs.io/en/latest/changelog.html#id63
        #
        # As long as we're running from the stack Python, we need to ensure
        # this is turned off.
        self._logger.debug("Forcing JUPYTER_PREFER_ENV_PATH to 'no'")
        self._env["JUPYTER_PREFER_ENV_PATH"] = "no"

    def _set_butler_credential_variables(self) -> None:
        # We split this up into environment manipulation and later
        # file substitution.  This is the environment part.
        self._logger.debug("Setting Butler credential variables")
        cred_dir = self._home / ".lsst"
        if "AWS_SHARED_CREDENTIALS_FILE" in self._env:
            awsname = Path(self._env["AWS_SHARED_CREDENTIALS_FILE"]).name
            self._env["ORIG_AWS_SHARED_CREDENTIALS_FILE"] = self._env[
                "AWS_SHARED_CREDENTIALS_FILE"
            ]
            newaws = str(cred_dir / awsname)
            self._env["AWS_SHARED_CREDENTIALS_FILE"] = newaws
            self._logger.debug(
                f"Set 'AWS_SHARED_CREDENTIALS_FILE' -> '{newaws}'"
            )
        if "PGPASSFILE" in self._env:
            pgpname = Path(self._env["PGPASSFILE"]).name
            newpg = str(cred_dir / pgpname)
            self._env["ORIG_PGPASSFILE"] = self._env["PGPASSFILE"]
            self._env["PGPASSFILE"] = newpg
            self._logger.debug(f"Set 'PGPASSFILE' -> '{newpg}'")

    #
    # The second big block tries to tidy the home directory.
    #
    def _tidy_homedir(self) -> None:
        self._clean_astropy_cache()
        self._test_for_space()

    def _clean_astropy_cache(self) -> None:
        # This is extremely conservative.  We only find URLs with an
        # "Expires" parameter (in practice, s3 signed URLs), and remove
        # the key and contents if the expiration is in the past.
        cachedir = self._home / ".astropy" / "cache" / "download" / "url"
        if not cachedir.exists():
            return
        candidates = [x for x in cachedir.iterdir() if x.is_dir()]
        for c in candidates:
            urlfile = c / "url"
            if not urlfile.is_file():
                continue
            try:
                url = urlfile.read_text()
            except Exception:
                self._logger.exception("Could not read {urlfile!s}")
                continue
            qry = urlparse(url).query
            if not qry:
                continue
            for key, value in parse_qsl(qry):
                if key.lower() == "expires":
                    self._handle_expiry(c, value)

    def _handle_expiry(self, cachefile: Path, expiry: str) -> None:
        try:
            exptime = int(expiry)
        except ValueError:
            self._logger.exception("Could not parse Expires header")
            return
        if time.time() > exptime:
            self._logger.debug(f"Removing expired cache {cachefile!s}")
            try:
                self._remove_astropy_cachedir(cachefile)
            except OSError:
                self._logger.exception(f"Failed to remove cache {cachefile!s}")
                # Having found the parameter, we are done with this url.
                return

    def _remove_astropy_cachedir(self, cachedir: Path) -> None:
        (cachedir / "url").unlink()
        (cachedir / "contents").unlink()
        cachedir.rmdir()

    def _test_for_space(self) -> None:
        cachefile = self._home / ".cache" / "1mb.txt"
        try:
            self._write_a_megabyte(cachefile)
        except OSError as exc:
            self._logger.warning("Could not write 1MB of text")
            self._set_abnormal_startup(exc)
        if self._broken:
            self._try_emergency_cleanup()
            try:
                # Did that clear enough room?
                self._write_a_megabyte(cachefile)
                self._clear_abnormal_startup()
            except OSError:
                pass  # Nope, stay broken.

    def _write_a_megabyte(self, cachefile: Path) -> None:
        # Try to write a 1M block, which should be enough to start the lab.
        sixteen = "0123456789abcdef"
        mega = sixteen * 64 * 1024

        parent = cachefile.parent
        parent.mkdir(exist_ok=True)
        cachefile.write_text(mega)
        self._remove_cachefile(cachefile)

    def _remove_cachefile(self, cachefile: Path) -> None:
        if cachefile.is_file():
            cachefile.unlink()

    def _try_emergency_cleanup(self) -> None:
        # We have either critically low space, or there's something else
        # wrong with the home directory.
        #
        # Try to reclaim the space by removing .cache and .astropy/cache.
        #
        # If we fail here, don't bother with recovery--we will be starting
        # in a degraded mode, and offering the user an explanation, anyway.
        self._logger.warning(
            "Attempting emergency cleanup of .cache and .astropy/cache"
        )
        try:
            for cdir in (
                (self._home / ".cache"),
                (self._home / ".astropy" / "cache"),
            ):
                shutil.rmtree(cdir, ignore_errors=True)
                cdir.mkdir(exist_ok=True)
        except Exception:
            self._logger.exception("Emergency cleanup failed")

    #
    # The third big block is a bunch of file manipulation.
    #
    def _copy_files_to_user_homedir(self) -> None:
        self._logger.debug("Copying files to user home directory")
        self._copy_butler_credentials()
        self._copy_logging_profile()
        self._copy_dircolors()
        self._copy_etc_skel()

    def _copy_butler_credentials(self) -> None:
        if "AWS_SHARED_CREDENTIALS_FILE" in self._env:
            self._merge_aws_creds()
        if "PGPASSFILE" in self._env:
            self._merge_pgpass()

    def _merge_aws_creds(self) -> None:
        #
        # Merge the config in the original credentials file and the one
        # in our homedir.  For any given section, we assume that the
        # information in the container ("original credentials files")
        # is correct, but leave any other user config alone.
        #
        ascf = "AWS_SHARED_CREDENTIALS_FILE"
        for ev in (ascf, "ORIG_" + ascf):
            if ev not in self._env:
                raise RSPStartupError(RSPErrorCode.EBADENV, None, ev)
        hc_path = Path(self._env["AWS_SHARED_CREDENTIALS_FILE"])
        if not hc_path.parent.exists():
            hc_path.parent.mkdir(mode=0o700, parents=True)
        hc_path.touch(mode=0o600, exist_ok=True)
        home_config = configparser.ConfigParser()
        home_config.read(str(hc_path))
        ro_config = configparser.ConfigParser()
        ro_config.read(self._env["ORIG_AWS_SHARED_CREDENTIALS_FILE"])
        for sect in ro_config.sections():
            home_config[sect] = ro_config[sect]
        with hc_path.open("w") as f:
            home_config.write(f)

    def _merge_pgpass(self) -> None:
        #
        # Same as above, but for pgpass files.
        #
        config = {}
        # Get current config from homedir
        ppf = "PGPASSFILE"
        for ev in (ppf, "ORIG_" + ppf):
            if ev not in self._env:
                raise RSPStartupError(RSPErrorCode.EBADENV)
        home_pgpass = Path(self._env["PGPASSFILE"])
        if not home_pgpass.parent.exists():
            home_pgpass.parent.mkdir(mode=0o700, parents=True)
        home_pgpass.touch(mode=0o600, exist_ok=True)
        lines = home_pgpass.read_text().splitlines()
        for line in lines:
            if ":" not in line:
                continue
            connection, passwd = line.rsplit(":", maxsplit=1)
            config[connection] = passwd.rstrip()
        # Update config from container-supplied one
        ro_pgpass = Path(self._env["ORIG_PGPASSFILE"])
        lines = ro_pgpass.read_text().splitlines()
        for line in lines:
            if ":" not in line:
                continue
            connection, passwd = line.rsplit(":", maxsplit=1)
            config[connection] = passwd.rstrip()
        with home_pgpass.open("w") as f:
            for connection in config:
                f.write(f"{connection}:{config[connection]}\n")

    def _copy_logging_profile(self) -> None:
        self._logger.debug("Copying logging profile if needed")
        user_profile = (
            self._home
            / ".ipython"
            / "profile_default"
            / "startup"
            / "20-logging.py"
        )
        #
        # We have a list of previous supplied versions of 20-logging.py.
        #
        # If the one we have has a hash that matches any of those, then
        # there is a new container-supplied 20-logging.py that should replace
        # it.  However, if we have a 20-logging.py that does not match
        # any of those, then it has been locally modified, and we should
        # not replace it.  If we don't have one at all, we need to copy it
        # into place.
        #
        copy = False
        if not user_profile.is_file():
            copy = True  # It doesn't exist, so we need one.
        else:
            user_loghash = hashlib.sha256(
                user_profile.read_bytes()
            ).hexdigest()
            if user_loghash in PREVIOUS_LOGGING_CHECKSUMS:
                self._logger.debug(
                    f"User log profile '{user_loghash}' is"
                    " out-of-date; replacing with current version."
                )
                copy = True
        if copy:
            pdir = user_profile.parent
            if not pdir.is_dir():
                pdir.mkdir(parents=True)
            jl_path = get_jupyterlab_config_dir()
            srcfile = jl_path / "etc" / "20-logging.py"
            # Location changed with two-python container.  Try each.
            if not srcfile.is_file():
                srcfile = jl_path / "20-logging.py"
            if not srcfile.is_file():
                self._logger.warning("Could not find source user log profile.")
                return
            user_profile.write_bytes(srcfile.read_bytes())

    def _copy_dircolors(self) -> None:
        self._logger.debug("Copying dircolors if needed")
        if not (self._home / ".dir_colors").exists():
            self._logger.debug("Copying dircolors")
            dc = ETC_PATH / "dircolors.ansi-universal"
            dc_txt = dc.read_text()
            (self._home / ".dir_colors").write_text(dc_txt)
        else:
            self._logger.debug("Copying dircolors not needed")

    def _copy_etc_skel(self) -> None:
        self._logger.debug("Copying files from /etc/skel if they don't exist")
        etc_skel = ETC_PATH / "skel"
        contents = etc_skel.walk()
        #
        # We assume that if the file exists at all, we should leave it alone.
        # Users are allowed to modify these, after all.
        #
        for entry in contents:
            dirpath = entry[0]
            dirs = [Path(x) for x in entry[1]]
            files = [Path(x) for x in entry[2]]
            # Determine what the destination directory should be
            if dirpath == etc_skel:
                current_dir = self._home
            else:
                current_dir = (
                    self._home / str(dirpath)[(len(str(etc_skel)) + 1) :]
                )
            # For each directory in the tree at this level:
            # if we don't already have one in our directory, make it.
            for d_item in dirs:
                if not (current_dir / d_item).is_dir():
                    (current_dir / d_item).mkdir()
                    self._logger.debug(f"Creating {current_dir / d_item!s}")
            # For each file in the tree at this level:
            # if we don't already have one in our directory, copy the
            # contents.
            for f_item in files:
                if not (current_dir / f_item).exists():
                    src = Path(entry[0] / f_item)
                    self._logger.debug(f"Creating {current_dir / f_item!s}")
                    src_contents = src.read_bytes()
                    (current_dir / f_item).write_bytes(src_contents)

    #
    # Now that we're not checking out notebooks anymore, all we have to do
    # with Git is install git-lfs.
    #
    def _setup_git(self) -> None:
        # Set up Git LFS
        self._setup_gitlfs()

    def _setup_gitlfs(self) -> None:
        # Check for git-lfs
        self._logger.debug("Installing Git LFS if needed")
        if not self._check_for_git_lfs():
            self._cmd.run("git", "lfs", "install")
            self._logger.debug("Git LFS installed")

    def _check_for_git_lfs(self) -> bool:
        gitconfig = self._home / ".gitconfig"
        if gitconfig.is_file():
            gc = gitconfig.read_text().splitlines()
            for line in gc:
                line.strip()
                if line == '[filter "lfs"]':
                    return True
        return False

    #
    # Start the lab.
    #
    def _launch(self) -> None:
        # We're about to start the lab: set the flag saying we're running
        # inside the lab.  It's used by shell startup.
        self._env["RUNNING_INSIDE_JUPYTERLAB"] = "TRUE"
        if bool(self._env.get("NONINTERACTIVE", "")):
            self._start_noninteractive()
            # We exec a lab; control never returns here
        self._modify_interactive_settings()
        self._start()

    def _modify_interactive_settings(self) -> None:
        self._logger.debug("Modifying interactive settings if needed")
        # These both write files; if either fails, start up but warn
        # the user their experience is likely to be bad.
        try:
            self._manage_access_token()
            self._increase_log_limit()
        except OSError as exc:
            self._set_abnormal_startup(exc)

    def _increase_log_limit(self) -> None:
        self._logger.debug("Increasing log limit if needed")
        settings: dict[str, Any] = {}
        settings_dir = (
            self._home
            / ".jupyter"
            / "lab"
            / "user-settings"
            / "@jupyterlab"
            / "notebook-extension"
        )
        settings_dir.mkdir(parents=True, exist_ok=True)
        settings_file = settings_dir / "tracker.jupyterlab.settings"
        if settings_file.is_file():
            with settings_file.open() as f:
                settings = json.load(f)
        current_limit = settings.get("maxNumberOutputs", 0)
        if current_limit < MAX_NUMBER_OUTPUTS:
            self._logger.warning(
                f"Changing maxNumberOutputs in {settings_file!s}"
                f" from {current_limit} to {MAX_NUMBER_OUTPUTS}"
            )
            settings["maxNumberOutputs"] = MAX_NUMBER_OUTPUTS
            with settings_file.open("w") as f:
                json.dump(settings, f, sort_keys=True, indent=4)
        else:
            self._logger.debug("Log limit increase not needed")

    def _manage_access_token(self) -> None:
        self._logger.debug("Updating access token")
        tokfile = self._home / ".access_token"
        tokfile.unlink(missing_ok=True)
        ctr_token = get_runtime_mounts_dir() / "secrets" / "token"
        if ctr_token.exists():
            self._logger.debug(f"Symlinking {tokfile!s}->{ctr_token!s}")
            tokfile.symlink_to(ctr_token)
            with contextlib.suppress(NotImplementedError):
                tokfile.chmod(0o600, follow_symlinks=False)
            return
        self._logger.debug("Did not find container token file")
        token = get_access_token()
        if token:
            tokfile.touch(mode=0o600)
            tokfile.write_text(token)
            self._logger.debug(f"Created {tokfile}")
        else:
            self._logger.debug("Could not determine access token")

    def _start_noninteractive(self) -> None:
        config_path = (
            get_runtime_mounts_dir()
            / "noninteractive"
            / "command"
            / "command.json"
        )
        launcher = NonInteractiveExecutor.from_config(config_path)
        launcher.execute(env=self._env)

    def _set_timeout_variables(self) -> list[str]:
        timeout_map = {
            "NO_ACTIVITY_TIMEOUT": "ServerApp.shutdown_no_activity_timeout",
            "CULL_KERNEL_IDLE_TIMEOUT": (
                "MappingKernelManager.cull_idle_timeout"
            ),
            "CULL_KERNEL_CONNECTED": "MappingKernelManager.cull_connected",
            "CULL_KERNEL_INTERVAL": "MappingKernelManager.cull_interval",
            "CULL_TERMINAL_INACTIVE_TIMEOUT": (
                "TerminalManager.cull_inactive_timeout"
            ),
            "CULL_TERMINAL_INTERVAL": "TerminalManager.cull_interval",
        }
        result: list[str] = []
        for setting in timeout_map:
            val = self._env.get(setting, "")
            if val:
                result.append(f"--{timeout_map[setting]}={val}")
        return result

    def _make_abnormal_startup_environment(self) -> None:
        # What we're doing is writing (we hope) someplace safe, be that
        # an empty, ephemeral filesystem (such as /tmp in any sanely-configured
        # K8s-based RSP) or in scratch space somewhere.
        #
        # Performance is irrelevant.  As we explain to the user, they should
        # not be using this lab for anything other than immediate problem
        # amelioration.

        # Try a sanity check and ensure that we are in fact in a broken state.
        if not self._broken:
            return

        txt = self._make_abnormal_landing_markdown()
        s_obj = {"defaultViewers": {"markdown": "Markdown Preview"}}
        s_txt = json.dumps(s_obj)

        try:
            temphome = self._env.get("SCRATCH_DIR", "/tmp")
            welcome = Path(temphome) / "notebooks" / "tutorials" / "welcome.md"
            welcome.parent.mkdir(exist_ok=True, parents=True)
            welcome.write_text(txt)
            settings = (
                Path(temphome)
                / ".jupyter"
                / "lab"
                / "user-settings"
                / "@jupyterlab"
                / "docmanager-extension"
                / "plugin.jupyterlab-settings"
            )
            settings.parent.mkdir(exist_ok=True, parents=True)
            settings.write_text(s_txt)
        except Exception:
            self._logger.exception(
                "Writing files to report abnormal startup failed"
            )

    def _make_abnormal_landing_markdown(self) -> str:
        user = self._env["USER"]
        home = self._env.get(
            "NUBLADO_HOME",
            self._env.get(
                "HOME",
                f"/home/{user}",  # Guess, albeit a good one.
            ),
        )

        errmsg = self._env.get("ABNORMAL_STARTUP_MESSAGE", "<no message>")
        errcode = self._env.get("ABNORMAL_STARTUP_ERRORCODE", "EUNKNOWN")

        self._logger.error(
            f"Abnormal startup: errorcode {errcode}; message {errmsg}"
        )

        open_an_issue = dedent(
            f"""

            Please open an issue with your RSP site administrator with the
            following information: `{errmsg}`
            """
        )

        # Start with generic error text.  It's very simple markdown, with a
        # heading and literal text only.

        txt = dedent("""
        # Abnormal startup

        Your Lab container did not start normally.

        Do not trust this lab for work you want to keep.

        """)

        # Now add error-specific advice.
        match errcode:
            case "EDQUOT":
                txt += dedent(
                    f"""
                    You have exceeded your quota.  Try using the terminal to
                    remove unneeded files in `{home}`.  You can use the
                    `quota` command to check your usage.

                    After that, shut down and restart the lab.  If that does
                    not result in a working lab:
                    """
                )
            case "ENOSPC":
                txt += dedent(
                    f"""
                    You have run out of filesystem space.  Try using the
                    terminal to remove unneeded files in `{home}`.  Since the
                    filesystem is full, this may not be something you can
                    correct.

                    After you have trimmed whatever possible, shut down and
                    restart the lab.

                    If that does not result in a working lab:
                    """
                )
            case "EROFS" | "EACCES":
                txt += dedent(
                    """
                    You do not have permission to write.  Ask your RSP
                    administrator to check ownership and permissions on your
                    directories.
                    """
                )
            case "EBADENV":
                txt += dedent(
                    """
                    You are missing environment variables necessary for RSP
                    operation.
                    """
                )
            case _:
                pass
        txt += dedent(open_an_issue)
        return txt

    def _start(self) -> None:
        log_level = "DEBUG" if self._debug else "INFO"
        notebook_dir = f"{self._home!s}"
        if self._broken:
            self._logger.warning(
                f"Abnormal startup: {self._env['ABNORMAL_STARTUP_MESSAGE']}"
            )
            self._make_abnormal_startup_environment()
            #
            # We will check to see if we got SCRATCH_DIR set before we broke,
            # and if so, use that, which would be a user-specific path on a
            # scratch filesystem.  If we didn't, we just use "/tmp" and hope
            # for the best.  Any reasonably-configured RSP running under K8s
            # will not have a shared "/tmp".
            #
            temphome = self._env.get("SCRATCH_DIR", "/tmp")
            self._logger.warning(f"Launching with homedir='{temphome}'")
            self._env["HOME"] = temphome
            os.environ["HOME"] = temphome
            notebook_dir = temphome

        cmd = [
            "jupyterhub-singleuser",
            "--ip=0.0.0.0",
            "--port=8888",
            "--no-browser",
            f"--notebook-dir={notebook_dir}",
            f"--log-level={log_level}",
            "--ContentsManager.allow_hidden=True",
            "--FileContentsManager.hide_globs=[]",
            "--KernelSpecManager.ensure_native_kernel=False",
            "--QtExporter.enabled=False",
            "--PDFExporter.enabled=False",
            "--WebPDFExporter.allow_chromium_download=True",
            "--MappingKernelManager.default_kernel_name=lsst",
            "--LabApp.check_for_updates_class=jupyterlab.NeverCheckForUpdate",
        ]
        cmd.extend(self._set_timeout_variables())
        self._logger.debug("Command to run:", command=cmd)
        # Flush open files before exec()
        sys.stdout.flush()
        sys.stderr.flush()
        # exec the Jupyter process directly.  We use os.execvpe()
        # because we have a list of arguments we just created and we
        # want to pass the environment we built up.
        os.execvpe(cmd[0], cmd, env=self._env)



"""Tests for startup object."""

import configparser
import errno
import json
import os
import shutil
from pathlib import Path

import pytest

import lsst.rsp
from lsst.rsp.startup.services.labrunner import LabRunner
from lsst.rsp.utils import get_jupyterlab_config_dir, get_runtime_mounts_dir


@pytest.mark.usefixtures("_rsp_env")
def test_object() -> None:
    lr = LabRunner()
    assert lr._debug is False


@pytest.mark.usefixtures("_rsp_env")
def test_debug_object(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("DEBUG", "1")
    lr = LabRunner()
    assert lr._debug is True


#
# Environment methods
#


@pytest.mark.usefixtures("_rsp_env")
def test_set_tmpdir(monkeypatch: pytest.MonkeyPatch) -> None:
    # Happy path.
    lr = LabRunner()
    lr._set_tmpdir_if_scratch_available()
    assert lr._env["TMPDIR"].endswith("/scratch/hambone/tmp")
    # Exists, but it's not a directory
    scratch_path = Path(lr._env["TMPDIR"])
    scratch_path.rmdir()
    scratch_path.touch()
    lr = LabRunner()
    lr._set_tmpdir_if_scratch_available()
    assert "TMPDIR" not in lr._env
    # Put it back the way it was
    scratch_path.unlink()
    # Pre-set TMPDIR.
    monkeypatch.setenv("TMPDIR", "/preset")
    lr = LabRunner()
    lr._set_tmpdir_if_scratch_available()
    assert lr._env["TMPDIR"] == "/preset"
    monkeypatch.delenv("TMPDIR")
    # Can't write to scratch dir
    monkeypatch.setenv("SCRATCH_PATH", "/nonexistent/scratch")
    lr = LabRunner()
    lr._set_tmpdir_if_scratch_available()
    assert "TMPDIR" not in lr._env
    monkeypatch.delenv("SCRATCH_PATH")
    scratch_path.parent.rmdir()


@pytest.mark.usefixtures("_rsp_env")
def test_set_butler_cache(monkeypatch: pytest.MonkeyPatch) -> None:
    env_v = "DAF_BUTLER_CACHE_DIRECTORY"
    # Happy path.
    lr = LabRunner()
    lr._set_butler_cache()
    assert lr._env[env_v].endswith("/scratch/hambone/butler_cache")
    dbc = Path(lr._env[env_v])
    dbc.rmdir()
    dbc.touch()
    lr = LabRunner()
    lr._set_butler_cache()
    assert lr._env[env_v] == "/tmp/butler_cache"
    # Put it back the way it was
    dbc.unlink()
    # Pre-set DAF_BUTLER_CACHE_DIR.
    monkeypatch.setenv(env_v, "/preset")
    lr = LabRunner()
    lr._set_butler_cache()
    assert lr._env[env_v] == "/preset"
    monkeypatch.delenv(env_v)
    dbc.parent.rmdir()


@pytest.mark.usefixtures("_rsp_env")
def test_cpu_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["CPU_LIMIT"] == "1"
    # We need a new LabRunner each time, because it reads its environment
    # only at __init()__
    monkeypatch.setenv("CPU_LIMIT", "NaN")
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["CPU_COUNT"] == "1"
    monkeypatch.setenv("CPU_LIMIT", "0.1")
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["GOTO_NUM_THREADS"] == "1"
    monkeypatch.setenv("CPU_LIMIT", "3.1")
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["MKL_DOMAIN_NUM_THREADS"] == "3"
    monkeypatch.setenv("CPU_LIMIT", "14")
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["MPI_NUM_THREADS"] == "14"


# No test for set_image_digest() because we test that in utils_test


@pytest.mark.usefixtures("_rsp_env")
def test_expand_panda_tilde(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == os.environ["HOME"]
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~hambone")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == os.environ["HOME"]
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~hambone/")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == os.environ["HOME"]
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~whoopsi")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == "~whoopsi"
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "/etc/panda")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == "/etc/panda"
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~/bar")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == str(
        Path(os.environ["HOME"]) / "bar"
    )


@pytest.mark.usefixtures("_rsp_env")
def test_set_timeout_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("NO_ACTIVITY_TIMEOUT", "300")
    lr = LabRunner()
    lr._set_timeout_variables()
    assert lr._env["NO_ACTIVITY_TIMEOUT"] == "300"
    assert "CULL_KERNEL_IDLE_TIMEOUT" not in lr._env


@pytest.mark.usefixtures("_rsp_env")
def test_set_firefly_variables(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("EXTERNAL_INSTANCE_URL", "https://lab.example.com:8443")
    lr = LabRunner()
    lr._set_firefly_variables()
    assert lr._env["FIREFLY_URL"] == "https://lab.example.com:8443/firefly"


@pytest.mark.usefixtures("_rsp_env")
def test_force_jupyter_prefer_env_path_false() -> None:
    lr = LabRunner()
    lr._force_jupyter_prefer_env_path_false()
    assert lr._env["JUPYTER_PREFER_ENV_PATH"] == "no"


@pytest.mark.usefixtures("_rsp_env")
def test_set_butler_credential_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("AWS_SHARED_CREDENTIALS_FILE", "/etc/secret/aws.creds")
    monkeypatch.setenv("PGPASSFILE", "/etc/secret/pgpass")
    lr = LabRunner()
    lr._set_butler_credential_variables()
    assert lr._env["AWS_SHARED_CREDENTIALS_FILE"] == str(
        lr._home / ".lsst" / "aws.creds"
    )
    assert (
        lr._env["ORIG_AWS_SHARED_CREDENTIALS_FILE"] == "/etc/secret/aws.creds"
    )
    assert lr._env["PGPASSFILE"] == str(lr._home / ".lsst" / "pgpass")
    assert lr._env["ORIG_PGPASSFILE"] == "/etc/secret/pgpass"


@pytest.mark.usefixtures("_rsp_env")
def test_busted_homedir(monkeypatch: pytest.MonkeyPatch) -> None:
    def out_of_space(lrobj: LabRunner, cachefile: Path) -> None:
        raise OSError(errno.EDQUOT, None, str(cachefile))

    monkeypatch.setattr(LabRunner, "_write_a_megabyte", out_of_space)
    lr = LabRunner()

    lr._test_for_space()

    assert lr._broken
    assert lr._env["ABNORMAL_STARTUP"] == "TRUE"
    assert lr._env["ABNORMAL_STARTUP_ERRNO"] == str(errno.EDQUOT)

    lr._clear_abnormal_startup()
    assert lr._broken is not True


#
# File manipulation tests
#


@pytest.mark.usefixtures("_rsp_env")
def test_create_credential_dir(monkeypatch: pytest.MonkeyPatch) -> None:
    secret_dir = get_runtime_mounts_dir() / "secrets"
    monkeypatch.setenv(
        "AWS_SHARED_CREDENTIALS_FILE", str(secret_dir / "aws-credentials.ini")
    )
    monkeypatch.setenv(
        "PGPASSFILE", str(secret_dir / "postgres-credentials.txt")
    )
    cred_dir = Path(os.environ["HOME"]) / ".lsst"
    assert cred_dir.exists()
    shutil.rmtree(cred_dir)
    assert not cred_dir.exists()
    lr = LabRunner()
    lr._set_butler_credential_variables()
    assert not cred_dir.exists()
    lr._copy_butler_credentials()
    assert cred_dir.exists()


@pytest.mark.usefixtures("_rsp_env")
def test_copy_butler_credentials(monkeypatch: pytest.MonkeyPatch) -> None:
    secret_dir = get_runtime_mounts_dir() / "secrets"
    monkeypatch.setenv(
        "AWS_SHARED_CREDENTIALS_FILE", str(secret_dir / "aws-credentials.ini")
    )
    monkeypatch.setenv(
        "PGPASSFILE", str(secret_dir / "postgres-credentials.txt")
    )
    lr = LabRunner()
    pg = lr._home / ".lsst" / "postgres-credentials.txt"
    lines = pg.read_text().splitlines()
    aws = lr._home / ".lsst" / "aws-credentials.ini"
    for line in lines:
        if line.startswith("127.0.0.1:5432:db01:postgres:"):
            assert line.rsplit(":", maxsplit=1)[1] == "gets_overwritten"
        if line.startswith("127.0.0.1:5532:db02:postgres:"):
            assert line.rsplit(":", maxsplit=1)[1] == "should_stay"
    cp = configparser.ConfigParser()
    cp.read(str(aws))
    assert set(cp.sections()) == {"default", "tertiary"}
    assert cp["default"]["aws_secret_access_key"] == "gets_overwritten"
    assert cp["tertiary"]["aws_secret_access_key"] == "key03"
    lr._set_butler_credential_variables()
    lr._copy_butler_credentials()
    lines = pg.read_text().splitlines()
    aws = lr._home / ".lsst" / "aws-credentials.ini"
    for line in lines:
        if line.startswith("127.0.0.1:5432:db01:postgres:"):
            assert line.rsplit(":", maxsplit=1)[1] == "s33kr1t"
        if line.startswith("127.0.0.1:5532:db02:postgres:"):
            assert line.rsplit(":", maxsplit=1)[1] == "should_stay"
    cp = configparser.ConfigParser()
    cp.read(str(aws))
    assert set(cp.sections()) == {"default", "secondary", "tertiary"}
    assert cp["default"]["aws_secret_access_key"] == "key01"
    assert cp["secondary"]["aws_secret_access_key"] == "key02"
    assert cp["tertiary"]["aws_secret_access_key"] == "key03"


@pytest.mark.usefixtures("_rsp_env")
def test_copy_logging_profile(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    pfile = (
        lr._home / ".ipython" / "profile_default" / "startup" / "20-logging.py"
    )
    assert not pfile.exists()
    pfile.parent.mkdir(parents=True)
    lr._copy_logging_profile()
    assert pfile.exists()
    h_contents = pfile.read_text()
    sfile = get_jupyterlab_config_dir() / "etc" / "20-logging.py"
    assert sfile.exists()
    s_contents = sfile.read_text()
    assert s_contents == h_contents
    h_contents += "\n# Locally modified\n"
    pfile.write_text(h_contents)
    lr._copy_logging_profile()
    new_contents = pfile.read_text()
    assert new_contents == h_contents
    assert new_contents != s_contents


@pytest.mark.usefixtures("_rsp_env")
def test_copy_dircolors(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    assert not (lr._home / ".dir_colors").exists()
    lr._copy_dircolors()
    assert (lr._home / ".dir_colors").exists()


@pytest.mark.usefixtures("_rsp_env")
def test_copy_etc_skel(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    assert not (lr._home / ".gitconfig").exists()
    assert not (lr._home / ".pythonrc").exists()
    etc = lsst.rsp.startup.constants.ETC_PATH
    prc = (etc / "skel" / ".pythonrc").read_text()
    prc += "\n# Local mods\n"
    (lr._home / ".pythonrc").write_text(prc)
    lr._copy_etc_skel()
    assert (lr._home / ".gitconfig").exists()
    sgc = (etc / "skel" / ".gitconfig").read_text()
    lgc = (lr._home / ".gitconfig").read_text()
    assert sgc == lgc
    src = (etc / "skel" / ".pythonrc").read_text()
    lrc = (lr._home / ".pythonrc").read_text()
    assert src != lrc
    assert (lr._home / "notebooks" / ".user_setups").exists()


@pytest.mark.usefixtures("_rsp_env")
def test_relocate_user_files(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("RESET_USER_ENV", "1")
    lr = LabRunner()
    assert not (lr._home / ".local").exists()
    assert not (lr._home / "notebooks" / ".user_setups").exists()
    (lr._home / ".local").mkdir()
    (lr._home / ".local" / "foo").write_text("bar")
    (lr._home / "notebooks").mkdir()
    (lr._home / "notebooks" / ".user_setups").write_text("#!/bin/sh\n")
    lr._relocate_user_environment_if_requested()
    assert not (lr._home / ".local").exists()
    assert not (lr._home / "notebooks" / ".user_setups").exists()
    reloc = next(iter((lr._home).glob(".user_env.*")))
    assert (reloc / "local" / "foo").read_text() == "bar"
    assert (reloc / "notebooks" / "user_setups").read_text() == "#!/bin/sh\n"


@pytest.mark.usefixtures("_rsp_env")
def test_setup_gitlfs(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    assert lr._check_for_git_lfs() is False
    lr._setup_gitlfs()
    assert lr._check_for_git_lfs() is True


#
# Interactive-mode-only tests
#


@pytest.mark.usefixtures("_rsp_env")
def test_increase_log_limit(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    settings = (
        lr._home
        / ".jupyter"
        / "lab"
        / "user-settings"
        / "@jupyterlab"
        / "notebook-extension"
        / "tracker.jupyterlab.settings"
    )
    assert not settings.exists()
    lr._increase_log_limit()
    assert settings.exists()
    with settings.open() as f:
        obj = json.load(f)
    assert obj["maxNumberOutputs"] >= 10000


@pytest.mark.usefixtures("_rsp_env")
def test_manage_access_token(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("DEBUG", "1")
    token = "token-of-esteem"
    monkeypatch.setenv("ACCESS_TOKEN", token)
    ctr_file = get_runtime_mounts_dir() / "secrets" / "token"
    # Save the token
    assert ctr_file.exists()
    save_token = ctr_file.read_text()
    # Remove the token file
    ctr_file.unlink()
    assert not ctr_file.exists()
    lr = LabRunner()
    tfile = lr._home / ".access_token"
    assert not tfile.exists()
    lr._manage_access_token()
    assert tfile.exists()
    assert tfile.read_text() == token
    tfile.unlink()
    ctr_file.write_text(token)
    assert ctr_file.exists()
    assert not tfile.exists()
    lr = LabRunner()
    lr._manage_access_token()
    assert tfile.exists()
    assert tfile.read_text() == token
    # Remove the rewritten saved file and replace with saved token.
    ctr_file.unlink()
    assert not ctr_file.exists()
    ctr_file.write_text(save_token)
    assert ctr_file.exists()



"""Test CST copier functionality."""

import json
import os
from pathlib import Path

import pytest

from lsst.rsp.startup.services.landing_page.exceptions import (
    DestinationError,
    DestinationIsDirectoryError,
    PrecheckError,
)
from lsst.rsp.startup.services.landing_page.provisioner import Provisioner


@pytest.mark.usefixtures("_init_container_fake_root")
def test_provisioner_basic() -> None:
    pr = Provisioner.from_env()

    fnames = [x.name for x in pr._source_files]
    outfiles = [pr._dest_dir / x for x in fnames]
    settings = (
        pr._home_dir
        / ".jupyter"
        / "lab"
        / "user-settings"
        / "@jupyterlab"
        / "docmanager-extension"
        / "plugin.jupyterlab-settings"
    )

    for outf in outfiles:
        assert not outf.exists()

    assert not settings.exists()

    pr.go()

    for outf in outfiles:
        assert outf.is_file()

    dest_files = list(pr._dest_dir.glob("*"))
    assert len(dest_files) == 1

    assert settings.is_file()
    s_obj = json.loads(settings.read_text())

    assert s_obj["defaultViewers"]["markdown"] == "Markdown Preview"


@pytest.mark.usefixtures("_init_container_fake_root")
def test_bad_source(monkeypatch: pytest.MonkeyPatch) -> None:
    srcdir = os.getenv("CST_LANDING_PAGE_SRC_DIR")
    assert srcdir is not None
    monkeypatch.setenv("CST_LANDING_PAGE_SRC_DIR", "/nonexistent")
    pr = Provisioner.from_env()
    with pytest.raises(PrecheckError):
        pr.go()
    monkeypatch.setenv("CST_LANDING_PAGE_SRC_DIR", srcdir)
    monkeypatch.setenv("CST_LANDING_PAGE_FILES", "nonexistent")
    pr = Provisioner.from_env()
    with pytest.raises(PrecheckError):
        pr.go()


def test_bad_homedir(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("NUBLADO_HOME", "/nonexistent")
    pr = Provisioner.from_env()
    with pytest.raises(PrecheckError):
        pr.go()


def test_bad_destination() -> None:
    pr = Provisioner.from_env()
    assert len(pr._source_files) != 0
    source_file = pr._source_files[0]
    destfile = pr._dest_dir / source_file.name

    if destfile.exists():
        # I mean technically you don't HAVE to run all the tests in order.
        # So you might not already have the file here.
        assert destfile.is_file()
        destfile.unlink()

    # Directory
    destfile.mkdir(parents=True, exist_ok=True)
    with pytest.raises(DestinationIsDirectoryError):
        pr.go()
    destfile.rmdir()

    # FIFO
    os.mkfifo(str(destfile), 0o600)
    with pytest.raises(DestinationError):
        pr.go()
    destfile.unlink()

    # Symlink
    destfile.symlink_to(Path("broken_link"))
    assert destfile.is_symlink()
    pr.go()
    assert destfile.is_file()

    # File requiring overwrite.
    bad_text = "!dlroW olleH\n"
    destfile.write_text(bad_text)
    source_text = source_file.read_text()
    assert source_text != bad_text
    assert destfile.is_file()
    pr.go()
    assert destfile.is_file()
    assert destfile.read_text() == source_text
    destfile.unlink()



"""Test the RSPClient."""

import pytest
from pytest_httpx import HTTPXMock

from lsst.rsp import get_query_history


@pytest.mark.usefixtures("_rsp_env")
@pytest.mark.asyncio
async def test_get_query_history(httpx_mock: HTTPXMock) -> None:
    """Ensure that get_query_history() works, which in turn will ensure
    that the RSPClient has the right headers and assembles its URL correctly.
    """
    httpx_mock.add_response(
        url="https://rsp.example.com/api/tap/async",
        match_headers={
            "Authorization": "Bearer gf-dummytoken",
            "Content-Type": "application/json",
        },
        text=(
            """<?xml version="1.0" encoding="UTF-8"?>
<uws:jobs xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
            xmlns:xlink="http://www.w3.org/1999/xlink" version="1.1">
  <uws:jobref id="phdl67i3tmklfdbz">
    <uws:phase>COMPLETED</uws:phase>
    <uws:runId>dp02_dc2_catalogs.Object - data-dev</uws:runId>
    <uws:ownerId>adam</uws:ownerId>
    <uws:creationTime>2025-01-15T23:36:17.931Z</uws:creationTime>
  </uws:jobref>
  <uws:jobref id="r4qyb04xesh7mbz3">
    <uws:phase>COMPLETED</uws:phase>
    <uws:ownerId>adam</uws:ownerId>
    <uws:creationTime>2024-12-05T17:49:27.518Z</uws:creationTime>
  </uws:jobref>
  <uws:jobref id="yk16agxjefl6gly6">
    <uws:phase>COMPLETED</uws:phase>
    <uws:runId>ivoa.ObsCore - data-dev</uws:runId>
    <uws:ownerId>adam</uws:ownerId>
    <uws:creationTime>2025-01-15T23:37:03.089Z</uws:creationTime>
  </uws:jobref>
</uws:jobs>"""
        ),
    )
    jobs = await get_query_history()
    assert jobs == ["phdl67i3tmklfdbz", "r4qyb04xesh7mbz3", "yk16agxjefl6gly6"]
    # The httpx mock will throw an error at teardown if we did not exercise
    # the mock, so we know the request matched both the URL and the headers.



"""CLI launchers for the Lab Runner and the landing page provisioner."""

import logging

from .services.labrunner import LabRunner
from .services.landing_page.provisioner import Provisioner


def launch_lab() -> None:
    """Make a LabRunner and call its single public method.  All settings are
    in the environment.
    """
    LabRunner().go()


def provision_landing_page() -> None:
    """Entry point for landing page provisioner.

    This is expected to be the entry point for an init container that sets
    up the landing page for the Nublado user's lab.

    Environment variable ``NUBLADO_HOME`` must be set.  This is
    Nublado-controller specific and set for Nublado init containers.

    We never want to raise an exception from here: if the container fails,
    we still want to start the user lab, but the experience might be
    suboptimal.
    """
    logger = logging.getLogger("landing_page_provisioner")
    logging.basicConfig()  # Not great but it'll do.
    try:
        provisioner = Provisioner.from_env()
        provisioner.go()
    except Exception:
        logger.exception("Provisioner failed")
